"""Functions for finding intersections of wellbore trajectories with surfaces and grids."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rqc
import resqpy.surface as rqs
import resqpy.grid_surface as rqgs
import resqpy.olio.intersection as meet
import resqpy.olio.box_utilities as bx
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec


def find_intersections_of_trajectory_with_surface(trajectory, surface):
    """Returns an array of triangle indices and an array of xyz of intersections of well trajectory with surface.

    arguments:
       trajectory (well.Trajectory object; or list thereof): the wellbore trajectory object(s) to find the intersections for;
          if a list of trajectories is provided then the return value is a corresponding list
       surface (surface.Surface object): the triangulated surface with which to intersect the trajectory

    returns:
       (numpy int array of shape (N,), numpy float array of shape (N, 3)): the first array is a list of surface
       triangle indices containing an intersection and the second array is the corresponding list of (x, y, z)
       intersection points; if the trajectory argument is a list of trajectories, then a correponding list of numpy
       array pairs is returned

    notes:
       interseections are found based on straight line segments between the trajectory control points, this will result
       in errors where there is significant curvature between neighbouring control points;
       a given triangle index might appear more than once in the first returned array
    """

    if isinstance(trajectory, list):
        trajectory_list = trajectory
    else:
        trajectory_list = [trajectory]

    t, p = surface.triangles_and_points()
    triangles = p[t]
    #   log.debug('finding intersections of wellbore trajectory(ies) with layer: ' + str(k0) + ' ' + ref_k_faces)
    results_list = []
    for traj in trajectory_list:
        all_intersects = meet.poly_line_triangles_intersects(traj.control_points, triangles)
        _, triangle_indices, intersect_points = meet.distilled_intersects(
            all_intersects)  # discard trajectory segment info
        if len(triangle_indices) == 0:
            results_list.append((None, None))
        else:
            results_list.append((triangle_indices, intersect_points))

    if isinstance(trajectory, list):
        return results_list
    return results_list[0]


def find_intersections_of_trajectory_with_layer_interface(trajectory,
                                                          grid,
                                                          k0 = 0,
                                                          ref_k_faces = 'top',
                                                          heal_faults = True,
                                                          quad_triangles = True):
    """Returns an array of column indices and an array of xyz of intersections of well trajectory with layer interface.

    arguments:
       trajectory (well.Trajectory object; or list thereof): the wellbore trajectory object(s) to find the intersections for;
          if a list of trajectories is provided then the return value is a corresponding list
       grid (grid.Grid object): the grid object from which a layer interface is to be converted to a surface
       k0 (int): the layer number (zero based) to be used
       ref_k_faces (string): either 'top' (the default) or 'base', indicating whether the top or the base
          interface of the layer is to be used
       heal_faults (boolean, default True): if True, faults will be 'healed' to give an untorn surface before looking
          for intersections; if False and the trajectory passes through a fault plane without intersecting the layer
          interface then no intersection will be identified; makes no difference if the grid is unfaulted
       quad_triangles (boolean, optional, default True): if True, 4 triangles are used to represent each cell k face,
          which gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution

    returns:
       (numpy int array of shape (N, 2), numpy float array of shape (N, 3)): the first array is a list of (j0, i0)
       indices of columns containing an intersection and the second array is the corresponding list of (x, y, z)
       intersection points; if the trajectory argument is a list of trajectories, then a correponding list of numpy
       array pairs is returned

    notes:
       interseections are found based on straight line segments between the trajectory control points, this will result
       in errors where there is significant curvature between neighbouring control points;
       a given (j0, i0) column might appear more than once in the first returned array
    """

    if isinstance(trajectory, list):
        trajectory_list = trajectory
    else:
        trajectory_list = [trajectory]

    log.debug('generating surface for layer: ' + str(k0) + ' ' + ref_k_faces)
    if grid.has_split_coordinate_lines and not heal_faults:
        surface = rqgs.generate_torn_surface_for_layer_interface(grid,
                                                                 k0 = k0,
                                                                 ref_k_faces = ref_k_faces,
                                                                 quad_triangles = quad_triangles)
    else:
        surface = rqgs.generate_untorn_surface_for_layer_interface(grid,
                                                                   k0 = k0,
                                                                   ref_k_faces = ref_k_faces,
                                                                   quad_triangles = quad_triangles)

    tri_intersect_list = find_intersections_of_trajectory_with_surface(trajectory_list, surface)

    results_list = []
    for triangle_indices, intersect_points in tri_intersect_list:
        if triangle_indices is None or intersect_points is None:
            results_list.append((None, None))
        else:
            j_list, i_list = surface.column_from_triangle_index(
                triangle_indices)  # todo: check this is valid for torn surfaces
            assert j_list is not None, 'failed to derive column indices from triangle indices'
            cols = np.stack((j_list, i_list), axis = -1)
            results_list.append((cols, intersect_points))

    if isinstance(trajectory, list):
        return results_list
    return results_list[0]


def find_first_intersection_of_trajectory_with_surface(trajectory,
                                                       surface,
                                                       start = 0,
                                                       start_xyz = None,
                                                       nudge = None,
                                                       return_second = False):
    """Returns xyz and other info of the first intersection of well trajectory with surface.

    arguments:
       trajectory (well.Trajectory object): the wellbore trajectory object(s) to find the intersection for
       surface (surface.Surface object): the triangulated surface with which to search for intersections
       start (int, default 0): an index into the trajectory knots to start the search from
       start_xyz (triple float, optional): if present, should lie on start segment and search starts from this point
       nudge (float, optional): if present and positive, starting xyz is nudged forward this distance along segment;
          if present and negative, starting xyz is nudged backward along segment
       return_second (boolean, default False): if True, a sextuplet is returned with the last 3 elements identifying
          the 'runner up' intersection in the same trajectory segment, or None, None, None if only one intersection found

    returns:
       a triplet if return_second is False; a sextuplet if return_second is True; the first triplet is:
       (numpy float array of shape (3,), int, int): being the (x, y, z) intersection point, and the trajectory segment number,
       and the triangle index of the first intersection point; or None, None, None if no intersection found;
       if return_second is True, the 4th, 5th & 6th return values are similar to the first three, conveying information
       about the second intersection of the same trajectory segment with the surface, or None, None, None if a no second
       intersection was found

    notes:
       interseections are found based on straight line segments between the trajectory control points, this will result
       in errors where there is significant curvature between neighbouring control points
    """

    if start >= trajectory.knot_count - 1:
        if return_second:
            return None, None, None, None, None, None
        return None, None, None

    t, p = surface.triangles_and_points()
    triangles = p[t]

    xyz = None
    tri = None
    knot = start
    xyz_2 = tri_2 = None
    if start_xyz is None:
        start_xyz = trajectory.control_points[knot]
    if knot < trajectory.knot_count - 2 and vec.isclose(start_xyz, trajectory.control_points[knot + 1]):
        knot += 1
    if nudge is not None and nudge != 0.0:
        remaining = trajectory.control_points[knot + 1] - start_xyz
        if vec.naive_length(remaining) <= nudge:
            start_xyz += 0.9 * remaining
        else:
            start_xyz += nudge * vec.unit_vector(remaining)
    while knot < trajectory.knot_count - 1:
        if start_xyz is None:
            line_p = trajectory.control_points[knot]
        else:
            line_p = start_xyz
        start_xyz = None
        line_v = trajectory.control_points[knot + 1] - line_p
        intersects = meet.line_triangles_intersects(line_p, line_v, triangles, line_segment = True)
        if not np.all(np.isnan(intersects)):
            intersects_indices = meet.intersects_indices(intersects)
            tri = intersects_indices[0]
            xyz = intersects[tri]
            if len(intersects_indices) > 1:
                best_manhattan = vec.manhattan_distance(line_p, xyz)
                runner_up = None
                for other_intersect_index_index in range(1, len(intersects_indices)):
                    other_tri = intersects_indices[other_intersect_index_index]
                    other_xyz = intersects[other_tri]
                    other_manhattan = vec.manhattan_distance(line_p, other_xyz)
                    if other_manhattan < best_manhattan:
                        if return_second:
                            tri_2 = tri
                            xyz_2 = xyz
                            runner_up = best_manhattan
                        tri = other_tri
                        xyz = other_xyz
                        best_manhattan = other_manhattan
                    elif return_second and (runner_up is None or other_manhattan < runner_up):
                        tri_2 = other_tri
                        xyz_2 = other_xyz
                        runner_up = other_manhattan
            break
        knot += 1
    if knot == trajectory.knot_count - 1:
        knot = None

    if return_second:
        return xyz, knot, tri, xyz_2, knot, tri_2
    return xyz, knot, tri


def find_first_intersection_of_trajectory_with_layer_interface(trajectory,
                                                               grid,
                                                               k0 = 0,
                                                               ref_k_faces = 'top',
                                                               start = 0,
                                                               heal_faults = False,
                                                               quad_triangles = True,
                                                               is_regular = False):
    """Returns info about the first intersection of well trajectory(s) with layer interface.

    arguments:
       trajectory (well.Trajectory object; or list thereof): the wellbore trajectory object(s) to find the intersection for;
          if a list of trajectories is provided then the return value is a corresponding list
       grid (grid.Grid object): the grid object from which a layer interface is to be converted to a surface
       k0 (int): the layer number (zero based) to be used
       ref_k_faces (string): either 'top' (the default) or 'base', indicating whether the top or the base
          interface of the layer is to be used
       start (int, default 0): an index into the trajectory knots to start the search from; is applied naively to all
          trajectories when a trajectory list is passed
       heal_faults (boolean, default False): if True, faults will be 'healed' to give an untorn surface before looking
          for intersections; if False and the trajectory passes through a fault plane without intersecting the layer
          interface then no intersection will be identified; makes no difference if the grid is unfaulted
       quad_triangles (boolean, optional, default True): if True, 4 triangles are used to represent each cell k face,
          which gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution
       is_regular (boolean, default False): set True if grid is a RegularGrid with IJK axes aligned with xyz axes

    returns:
       (numpy float array of shape (3,), int, (int, int)): being the (x, y, z) intersection point, and the trajectory segment number,
       and the (j0, i0) column number of the first intersection point;
       or None, None, (None, None) if no intersection found;
       if the trajectory argument is a list of trajectories, then correponding list of numpy array, list of int, list of int pair
       are returned

    notes:
       interseections are found based on straight line segments between the trajectory control points, this will result
       in errors where there is significant curvature between neighbouring control points
    """

    if isinstance(trajectory, list):
        trajectory_list = trajectory
    else:
        trajectory_list = [trajectory]

    # log.debug('generating surface for layer: ' + str(k0) + ' ' + ref_k_faces)
    if is_regular:
        interface_depth = grid.block_dxyz_dkji[0, 2] * (k0 + (1 if ref_k_faces == 'base' else 0))
        surface = rqs.Surface(grid.model)
        surface.set_to_horizontal_plane(interface_depth, grid.xyz_box())
    elif grid.has_split_coordinate_lines and not heal_faults:
        surface = rqgs.generate_torn_surface_for_layer_interface(grid,
                                                                 k0 = k0,
                                                                 ref_k_faces = ref_k_faces,
                                                                 quad_triangles = quad_triangles)
    else:
        surface = rqgs.generate_untorn_surface_for_layer_interface(grid,
                                                                   k0 = k0,
                                                                   ref_k_faces = ref_k_faces,
                                                                   quad_triangles = quad_triangles)

    # log.debug('finding intersections of wellbore trajectory(ies) with layer: ' + str(k0) + ' ' + ref_k_faces)
    results_list = []
    segment_list = []
    col_list = []
    for traj in trajectory_list:
        xyz, knot, tri = find_first_intersection_of_trajectory_with_surface(traj, surface, start = start)
        if is_regular:
            col_j = int(xyz[1] / grid.block_dxyz_dkji[1, 1])
            col_i = int(xyz[1] / grid.block_dxyz_dkji[2, 0])
            assert 0 <= col_i < grid.ni and 0 <= col_j < grid.nj
            col = (col_j, col_i)
        else:
            col = surface.column_from_triangle_index(tri)  # j, i pair returned
        results_list.append(xyz)
        segment_list.append(knot)
        col_list.append(col)

    if isinstance(trajectory, list):
        return results_list, segment_list, col_list
    return results_list[0], segment_list[0], col_list[0]


def find_first_intersection_of_trajectory_with_cell_surface(trajectory,
                                                            grid,
                                                            kji0,
                                                            start_knot,
                                                            start_xyz = None,
                                                            nudge = 0.001,
                                                            quad_triangles = True):
    """Return first intersection with cell's surface found along a trajectory."""

    cp = grid.corner_points(kji0)
    cell_surface = rqs.Surface(grid.model)
    cell_surface.set_to_single_cell_faces_from_corner_points(cp, quad_triangles = quad_triangles)
    t, p = cell_surface.triangles_and_points()
    triangles = p[t]
    knot = start_knot
    xyz = None
    axis = polarity = None
    if start_xyz is not None:
        start_xyz = np.array(start_xyz)
    while knot < trajectory.knot_count - 1:
        if knot == start_knot and start_xyz is not None:
            if knot < trajectory.knot_count - 2 and vec.isclose(start_xyz, trajectory.control_points[knot + 1]):
                knot += 1
            if nudge is not None and nudge != 0.0:
                remaining = trajectory.control_points[knot + 1] - start_xyz
                if vec.naive_length(remaining) <= nudge:
                    start_xyz += 0.9 * remaining
                else:
                    start_xyz += nudge * vec.unit_vector(remaining)
            line_p = start_xyz
        else:
            line_p = trajectory.control_points[knot]
        line_v = trajectory.control_points[knot + 1] - line_p
        #      log.debug('kji0: ' + str(kji0) + '; knot: ' + str(knot) + '; line p: ' + str(line_p) + '; v: ' + str(line_v))
        intersects = meet.line_triangles_intersects(line_p, line_v, triangles, line_segment = True)
        if not np.all(np.isnan(intersects)):
            # if more than one intersect, could find one closest to line_p; should not be needed when starting inside cell
            tri = meet.intersects_indices(intersects)[0]
            xyz = intersects[tri]
            _, axis, polarity = cell_surface.cell_axis_and_polarity_from_triangle_index(tri)
            # log.debug('intersection xyz: ' + str(xyz) + '; tri: ' + str(tri) + '; axis: ' + str(axis) + '; polarity: ' +
            #           str(polarity))
            break
        knot += 1

    if knot == trajectory.knot_count - 1:
        knot = None
    return xyz, knot, axis, polarity


def find_intersection_of_trajectory_interval_with_column_face(trajectory,
                                                              grid,
                                                              start_knot,
                                                              col_ji0,
                                                              axis,
                                                              polarity,
                                                              start_xyz = None,
                                                              nudge = None,
                                                              quad_triangles = True):
    """Searches for intersection of a single trajectory segment with an I or J column face.

    returns:
       xyz, k0

    note:
       does not support k gaps
    """

    # build a set of faces for the column face
    # extract column face points into a shape ready to become a mesh
    _, col_face_surface = rqgs.create_column_face_mesh_and_surface(grid,
                                                                   col_ji0,
                                                                   axis,
                                                                   polarity,
                                                                   quad_triangles = quad_triangles)
    t, p = col_face_surface.triangles_and_points()
    triangles = p[t]
    log.debug(f"intersecting trajectory segment with column face ji0 {col_ji0} face {'KJI'[axis]}{'-+'[polarity]}")
    if start_xyz is not None:
        if nudge is not None and nudge != 0.0:  # here we typically nudge back up the wellbore
            start_xyz += nudge * vec.unit_vector(trajectory.control_points[start_knot + 1] - start_xyz)
        line_p = start_xyz
    else:
        line_p = trajectory.control_points[start_knot]
    line_v = trajectory.control_points[start_knot + 1] - line_p
    intersects = meet.line_triangles_intersects(line_p, line_v, triangles, line_segment = True)
    if np.all(np.isnan(intersects)):
        return None, None
    tri = meet.intersects_indices(intersects)[0]  # todo: if more than one, find one closest to line_p
    xyz = intersects[tri]
    _, k0 = col_face_surface.column_from_triangle_index(tri)  # 'column' in method name here maps to layers!
    return xyz, k0


def trajectory_grid_overlap(trajectory, grid, lazy = False):
    """Returns True if there is some overlap of the xyz boxes for the trajectory and grid, False otherwise.

    notes:
       overlap of the xyz boxes does not guarantee that the trajectory intersects the grid;
       a return value of False guarantees that the trajectory does not intersect the grid
    """

    traj_box = np.empty((2, 3))
    traj_box[0] = np.amin(trajectory.control_points, axis = 0)
    traj_box[1] = np.amax(trajectory.control_points, axis = 0)
    grid_box = grid.xyz_box(lazy = lazy, local = True)
    if not bu.matching_uuids(trajectory.crs_uuid, grid.crs_uuid):
        t_crs = rqc.Crs(trajectory.model, uuid = trajectory.crs_uuid)
        g_crs = rqc.Crs(grid.model, uuid = grid.crs_uuid)
        t_crs.convert_array_to(g_crs, traj_box)
        t_box = traj_box.copy()
        traj_box[0] = np.min(t_box, axis = 0)
        traj_box[1] = np.max(t_box, axis = 0)
    log.debug(f'overlap check: traj: {traj_box}; grid{grid_box}; overlap: {bx.boxes_overlap(traj_box, grid_box)}')
    return bx.boxes_overlap(traj_box, grid_box)



"""Multiprocessing wrapper functions for the Mesh class."""

import logging

log = logging.getLogger(__name__)

import uuid
from typing import Tuple, Union, List
from pathlib import Path
from uuid import UUID

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.surface as rqs
import resqpy.multi_processing as rqmp


def mesh_from_regular_grid_column_property_wrapper(
    index: int,
    parent_tmp_dir: str,
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    prop_uuids: List[Union[UUID, str]],
) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Multiprocessing wrapper function of the Mesh from_regular_grid_column_property method.

    arguments:
        index (int): the index of the function call from the multiprocessing function
        parent_tmp_dir (str): the parent temporary directory path from the multiprocessing function
        grid_epc (str): epc file path where the grid is saved
        grid_uuid (UUID or str): UUID (universally unique identifier) of the regular grid object
        prop_uuids (list of UUID or str): a list of the property uuids used to create each Mesh
            and their relationship

    returns:
        Tuple containing:
            - index (int): the index passed to the function
            - success (bool): True if all the Mesh objects could be created, False otherwise
            - epc_file (str): the epc file path where the objects are stored
            - uuid_list (List[UUID/str]): list of UUIDs of relevant objects

    notes:
        Use this function as argument to multiprocessing function; it will create a new model that
        is saved in a temporary epc file and returns the required values, which are used in the
        multiprocessing function to recombine all the objects into a single epc file; applications
        should generally call mesh_from_regular_grid_column_property_batch() which makes use of
        this wrapper
    """
    uuid_list = []
    tmp_dir = Path(parent_tmp_dir) / f"{uuid.uuid4()}"
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = str(tmp_dir / "wrapper.epc")
    model = rq.new_model(epc_file = epc_file, quiet = True)

    g_model = rq.Model(grid_epc, quiet = True)
    g_crs_uuid = g_model.uuid(obj_type = "LocalDepth3dCrs",
                              related_uuid = grid_uuid)  # todo: check this relationship exists

    if g_crs_uuid is not None:
        model.copy_uuid_from_other_model(g_model, g_crs_uuid)
        uuid_list.append(g_crs_uuid)
    model.copy_uuid_from_other_model(g_model, uuid = grid_uuid)

    for prop_uuid in prop_uuids:
        model.copy_uuid_from_other_model(g_model, uuid = prop_uuid)

    grid = grr.RegularGrid(parent_model = model, uuid = grid_uuid)

    success = True
    for prop_uuid in prop_uuids:
        mesh = rqs.Mesh.from_regular_grid_column_property(model, grid.uuid, prop_uuid)
        if mesh is None:
            success = False
            continue
        mesh.write_hdf5()
        mesh.create_xml()
        uuid_list.append(mesh.uuid)
        model.create_reciprocal_relationship_uuids(mesh.uuid, "sourceObject", prop_uuid, "detinationObject")

    model.store_epc(quiet = True)

    return index, success, epc_file, uuid_list


def mesh_from_regular_grid_column_property_batch(grid_epc: str,
                                                 grid_uuid: Union[UUID, str],
                                                 prop_uuids: List[Union[UUID, str]],
                                                 recombined_epc: str,
                                                 cluster,
                                                 n_workers: int,
                                                 require_success: bool = False,
                                                 tmp_dir_path: Union[Path, str] = '.') -> List[bool]:
    """Creates Mesh objects from a list of property uuids in parallel.

    arguments:
        grid_epc (str): epc file path where the grid is saved
        grid_uuid (UUID/str): UUID (universally unique identifier) of the grid object
        prop_uuids (List[UUID/str]): a list of the column property uuids used to create each Mesh
            and their relationship
        recombined_epc (Path/str): A pathlib Path or path string of
            where the combined epc will be saved
        cluster (LocalCluster/JobQueueCluster): a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc
        n_workers (int): the number of workers on the cluster
        require_success (bool, default False): if True an exception is raised if any failures
        tmp_dir_path (str or Path, default '.'): the directory within which temporary directories will reside

    returns:
        success_list (list of bool): A boolean list of successful function calls; one value per batch
    """
    n_uuids = len(prop_uuids)
    prop_uuids_list = [prop_uuids[i * n_uuids // n_workers:(i + 1) * n_uuids // n_workers] for i in range(n_workers)]

    kwargs_list = []
    for prop_uuids in prop_uuids_list:
        d = {
            "grid_epc": grid_epc,
            "grid_uuid": grid_uuid,
            "prop_uuids": prop_uuids,
        }
        kwargs_list.append(d)

    success_list = rqmp.function_multiprocessing(mesh_from_regular_grid_column_property_wrapper,
                                                 kwargs_list,
                                                 recombined_epc,
                                                 cluster,
                                                 require_success = require_success,
                                                 tmp_dir_path = tmp_dir_path)

    return success_list



"""Functions to load data from various ASCII simulator file formats."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.olio.ab_toolbox as abt
import resqpy.olio.box_utilities as box
import resqpy.olio.grid_functions as gf
import resqpy.olio.keyword_files as kf
import resqpy.olio.write_data as wd


def file_exists(file_name, must_be_more_recent_than_file = None):
    """Returns True if the file exists (and is more recent than other file, if given)."""

    if file_name is None or len(file_name) == 0:
        return False
    existe = os.path.exists(file_name)
    if not existe:
        return False
    if not must_be_more_recent_than_file or file_name == must_be_more_recent_than_file or \
            not os.path.exists(must_be_more_recent_than_file):
        return True
    return os.path.getmtime(file_name) > os.path.getmtime(must_be_more_recent_than_file)


######################################################################################################
# load_corp_array_from_file():
# function to load a nexus corp array of data from a named file
# if the grid extent is not known, the file must be free of comments
# NB: extent_kji here is extent of grid, rather than that of the corp array
# returns a pagoda style 7D array, resequenced


def load_corp_array_from_file(file_name,
                              extent_kji = None,
                              corp_bin = False,
                              swap_bytes = True,
                              max_lines_for_keyword = 100,
                              comment_char = None,
                              data_free_of_comments = False,
                              use_binary = False,
                              eight_mode = False,
                              use_numbers_only = None):
    """Loads a nexus corner point (CORP) array from a file, returns a 7D numpy array in pagoda ordering.

    arguments:
       file_name: The name of an ascii file holding the CORP data (no other keywords with numeric data
                should be in the file); write access to the directory is likely to be needed if
                use_binary is True
       extent_kji: The extent of the grid as a list or a 3 element numpy array, in the order [NK, NJ, NI].
                If extent_kji is None, the extent is figured out from the data.  It must be
                given for 1D or 2D models
       corp_bin (boolean, default False): if True, input file is in bespoke corp binary format, otherwise ascii
       swap_bytes (boolean, default True): if True, byte ordering of corp bin data is reversed; only relevant
                if corp_bin is True
       max_lines_for_keyword: the maximum number of lines to search for CORP keyword; set to zero if file is
                known to be data only
       comment_char: A single character string which is interpreted as introducing a comment
       data_free_of_comments: If True, once the numeric data is encountered, it is assumed that there are no
                further comments (allowing a faster load)
       use_binary: If True, a more recent file containing a pure binary copy of the data is looked for first,
                in the same directory; if found, the data is loaded directly from that file; if not found, the
                binary file is created after the ascii has been loaded (ready for next time)
       eight_mode: If True, the data is assumed to be in CORP EIGHT ordering; otherwise the normal ordering
                (The code does not look for keywords.); this is not automatically determined from any keyword
                in the file
       use_numbers_only: no longer in use, ignored

    returns:
       A numpy array containing the CORP data in 7D pagoda protocol ordering. The extent of the grid, and hence
       shape of the array is determined from the corner point data unless extent_kji has been specified.
    """

    #   assert(extent_kji is not None or data_free_of_comments)   # no longer a requirement

    if extent_kji is None:
        extent = None
    else:
        extent = [extent_kji[0] * extent_kji[1] * extent_kji[2], 8, 3]

    if corp_bin:
        bin_file_size = os.path.getsize(file_name)
        bin_cell_count, remainder = divmod(bin_file_size,
                                           26 * 4)  # corp bin format has records of 24 + 2 (head, tail) 32 bit words
        if remainder:
            log.error('corp binary file ' + str(file_name) + ' is not a whole number of records')
            return None
        dt = np.dtype('float32')
        if swap_bytes:
            dt = dt.newbyteorder()
        cp_array = np.fromfile(file_name, dtype = dt, count = 26 * bin_cell_count).reshape((-1, 26))[:, 1:25]
        if extent_kji is not None:
            if bin_cell_count != (extent_kji[0] * extent_kji[1] * extent_kji[2]):
                log.error('corp binary file ' + str(file_name) + ' contains data for ' + str(bin_cell_count) +
                          ' cells; extent requires ' + str(extent_kji[0] * extent_kji[1] * extent_kji[2]))

    else:
        cp_array = load_array_from_file(file_name,
                                        extent = extent,
                                        data_type = 'real',
                                        keyword = 'CORP',
                                        max_lines_for_keyword = max_lines_for_keyword,
                                        comment_char = comment_char,
                                        data_free_of_comments = data_free_of_comments,
                                        use_binary = use_binary)

    cell_count, remainder = divmod(cp_array.size, 24)
    if remainder:
        log.error('file ' + file_name + ' contains ' + str(cp_array.size) + ' data, which is not a multiple of 24')
        return None

    cp_array = cp_array.reshape(1, 1, cell_count, 2, 2, 2, 3)  # pagoda 7D, temporarily with all cells on I axis
    log.debug('resequencing corner point data')
    gf.resequence_nexus_corp(cp_array, eight_mode = eight_mode)  # switches IP points where JP = 1 (unless eight_mode)

    if extent_kji is None:
        log.info('determining grid extent from corner points')
        extent_kji = gf.determine_corp_extent(cp_array)
        if extent_kji is None:
            log.error('failed to determine extent of grid from corner points')
            return cp_array

    return cp_array.reshape(extent_kji[0], extent_kji[1], extent_kji[2], 2, 2, 2, 3)


######################################################################################################
# load_array_from_file():
# function to load an array of data from a named file
# if the extent is not known (None), the file must be free of comments
# a new numpy array is returned
# binary data file creation & reuse is supported


def load_array_from_file(file_name,
                         extent = None,
                         data_type = 'real',
                         keyword = None,
                         max_lines_for_keyword = None,
                         comment_char = None,
                         data_free_of_comments = False,
                         use_binary = False,
                         use_numbers_only = None):
    """Load an array from an ascii (or pure binary) file.

    Arguments are similar to those for load_corp_array_from_file().
    """

    if not use_binary:
        return load_array_from_ascii_file(file_name,
                                          extent = extent,
                                          data_type = data_type,
                                          keyword = keyword,
                                          max_lines_for_keyword = max_lines_for_keyword,
                                          comment_char = comment_char,
                                          data_free_of_comments = data_free_of_comments)

    (extension, ab_type) = abt.binary_file_extension_and_np_type_for_data_type(data_type)

    if extent is None:
        cell_count = -1  # np.fromfile interprets this as 'read everything'
        log.debug('Loading unknown number of array data elements from file ' + file_name)
    else:
        cell_count = np.prod(extent)
        log.debug('Loading %1d array data elements from file %s', cell_count, file_name)

    ascii_file_name = file_name
    binary_file_name = file_name
    if len(binary_file_name) < 4 or binary_file_name[-3:] != extension:
        binary_file_name += extension
    else:
        ascii_file_name = ascii_file_name[:-3]  # strip off '.db' or similar

    try:  # tentatively try to read data from an existing binary file, if present
        if file_exists(binary_file_name, must_be_more_recent_than_file = ascii_file_name):
            with open(binary_file_name, 'rb') as binary_file_in:
                result = np.fromfile(binary_file_in, dtype = ab_type, count = cell_count)
                if extent is not None:
                    result = result.reshape(extent)
                # check that end of file has been reached, ie. not too much data in file
                try:  # expected to return null
                    c = binary_file_in.read(1)
                    if len(c):
                        log.warning('binary file contains more data than expected: ' + binary_file_name)
                except Exception:
                    pass
                log.info('Data loaded from binary file %s', binary_file_name)
                return result
    except Exception:
        pass

    # read from ascii file
    result = load_array_from_ascii_file(file_name,
                                        extent = extent,
                                        data_type = data_type,
                                        keyword = keyword,
                                        max_lines_for_keyword = max_lines_for_keyword,
                                        comment_char = comment_char,
                                        data_free_of_comments = data_free_of_comments)

    # create a binary file (to be used next time)
    try:
        wd.write_pure_binary_data(binary_file_name, result)
    except Exception:
        log.warn('Failed to write data to binary file %s', binary_file_name)
        # todo: could delete the binary file in case a corrupt file is left for use next time

    return result


# end of load_array_from_file() def
######################################################################################################

######################################################################################################
# load_array_from_ascii_file():
# function to load an array of data from a named ascii file
# if the extent is not known (None), the file must be free of comments
# a new numpy array is returned


def load_array_from_ascii_file(file_name,
                               extent = None,
                               data_type = 'real',
                               keyword = None,
                               max_lines_for_keyword = None,
                               comment_char = None,
                               data_free_of_comments = False,
                               skip_c_space = True,
                               use_numbers_only = None):
    """Returns a numpy array with data loaded from an ascii file.

    arguments:
       file_name: string holding name of the existing ascii data file, eg. 'Cell_depth.dat'
       extent: a python list of integers specifying the extent (shape) of the array, eg. [148, 270, 103];
          if None, all data is read and returned as a 'flat' 1D array (data must be free from comments)
       data_type: the type of individual data elements, one of 'real','float','int', 'integer', 'bool' or 'boolean'
       keyword: if present, an attempt is made to find the keyword before reading data, if keyword is None or is
          not found, data is read from the start of the file
       max_lines_for_keyword: can be used to limit the search for keyword (for speed efficiency)
       comment_char: a single character string being the character used to introduce a comment in the file
       data_free_of_comments: if set to True, a faster load is used once any header line comments have been
          skipped
       skip_c_space: if True then a line starting 'C' followed by white space is skipped as a comment
       use_numbers_only: this argument is no longer in use and is ignored

    returns:
       a numpy array of shape specified in extent argument with dtype matching data_type

    example call:
       depth_array = load_data.load_array_from_ascii_file('Cell_depth.dat', [148, 270, 103])

    notes:
       In all use cases, this function is designed to load a single array of data from an ascii file
       that DOES NOT CONTAIN OTHER ARRAYS as well, ie. data for a single simulation keyword in the file.

       If skip_c_space is True, lines starting 'C ' are
       also treated as comments.  If data_free_of_comments is True, there must be
       at least one blank line before the data begins, and no further comments are permitted.
       (This format is designed to handle data files generated by a commonly used geomodelling package.)

       Repeat counts must not be present in the ascii data.

       The extent, if present, can contain any number of dimensions, typically 3 for reservoir modelling work.
       The total number of numbers in the file must match the number of elements in the given extent
       (ie. the product of the list of numbers in the extent argument).
       The order of indices in extent should be 'slowest changing' first, eg.: k,j,i

       The data_type defaults to 'real'
       'real' and 'float' are synonymous; 'int' and 'integer' are synonymous;
       'bool' and 'boolean' are synonymous; default is 'real'
       The numpy data type will be the default 64 bit float or 64 bit int
    """
    # todo: Code enhancement could cater for 32 bit options (and 8 bit for bool) if needed to reduce memory usage

    if extent is None:
        cell_count = -1  # np.fromfile interprets this as 'read everything'
        log.debug('Loading unknown number of array data elements from ascii file ' + file_name)
    else:
        cell_count = np.prod(extent)
        log.debug('Loading %1d array data elements from ascii file %s', cell_count, file_name)

    if data_type in ['real', 'float', float]:
        d_type = 'float'
    elif data_type in ['int', 'integer', int]:
        d_type = 'int'
    elif data_type in ['bool', 'boolean', bool]:
        d_type = 'int'  # read booleans as 0 or 1 and convert after
    else:
        assert False, 'Unknown data_type passed to load_array_from_ascii_file' + str(data_type)

    read_file_name = file_name

    with open(read_file_name, 'r') as data_file:

        if not comment_char and not data_free_of_comments:
            comment_char = kf.guess_comment_char(data_file)
            if not comment_char:
                comment_char = '!'

        if keyword:
            keyword_found = kf.find_keyword(data_file, keyword, max_lines = max_lines_for_keyword)
            if keyword_found:
                data_file.readline()  # skip keyword line

        kf.skip_blank_lines_and_comments(data_file, comment_char = comment_char, skip_c_space = skip_c_space)

        result = None

        if data_free_of_comments:  # use numpy fromfile function after passing header comments

            result = np.fromfile(data_file, dtype = d_type, count = cell_count, sep = ' ')

            if result is None:
                data_file.seek(0)

        if result is None:

            # builds one very big string, stripping trailing comments

            #         s = ''
            #         while True:
            #            r = data_file.readline()
            #            if len(r) == 0: break
            #            s += r.partition(comment_char)[0] + ' '
            #         result = np.fromstring(s, dtype = d_type, count = cell_count, sep = ' ')
            #         # note: extra data will go unnoticed if cell count known!
            #         del s

            b = bytearray(data_file.read().encode())
            bc = comment_char.encode()
            nl = b'\n'
            sp = b' '
            i = 0
            while True:
                i = b.find(bc, i)
                if i < 0:
                    break
                eol = b.find(nl, i)
                if eol < 0:
                    eol = len(b)
                b[i:eol] = sp * (eol - i)

            result = np.fromstring(b.decode(), dtype = d_type, count = cell_count, sep = ' ')
            # note: extra data will go unnoticed if cell count known!
            del b

        if result is None:

            assert (extent is not None)  # TODO: remove this restriction by dynamically extending a flat array

            view_1D = np.zeros(extent, dtype = d_type).flatten()
            elements = view_1D.size
            start_of_line = True

            for index in range(elements):

                while True:
                    ch = data_file.read(1)
                    assert (ch != '')  # premature end of file
                    if ch == comment_char:
                        data_file.readline()
                        start_of_line = True
                        continue
                    if skip_c_space and start_of_line and ch in ['C', 'c']:
                        next_ch = data_file.read(1)
                        if next_ch in ' \t\n':
                            if next_ch != '\n':
                                data_file.readline()
                            continue
                        ch += next_ch
                        break
                    if ch not in ' \t\n':
                        break

                start_of_line = False
                word = ch
                while True:
                    ch = data_file.read(1)
                    if ch == '':  # end of file (this assumes a whitespace character after last datum)
                        log.error('Not enough data in file %s: %1d of %1d numbers read', file_name, index, elements)
                        assert False, 'not enough data in file'
                    if ch in ' \t\n':
                        break
                    word += ch
                if ch == '\n':
                    start_of_line = True
                if d_type == 'float':
                    view_1D[index] = float(word)
                else:
                    view_1D[index] = int(word)

            result = view_1D

        kf.skip_blank_lines_and_comments(data_file, comment_char = comment_char, skip_c_space = skip_c_space)
        if not kf.end_of_file(data_file):
            assert False, 'too much data in ascii file: ' + file_name

    if result is not None and extent is not None:
        result = result.reshape(extent)

    if data_type in ['bool', 'boolean', bool]:
        return result != 0  # convert to boolean
    else:
        return result


# end of load_array_from_ascii_file() def
######################################################################################################



"""High level add edges per column property array function."""

import resqpy.derived_model
import resqpy.property as rqp

import resqpy.derived_model._add_one_grid_property_array as rqdm_aogp


def add_edges_per_column_property_array(epc_file,
                                        a,
                                        property_kind,
                                        grid_uuid = None,
                                        source_info = 'imported',
                                        title = None,
                                        discrete = False,
                                        uom = None,
                                        time_index = None,
                                        time_series_uuid = None,
                                        string_lookup_uuid = None,
                                        null_value = None,
                                        facet_type = None,
                                        facet = None,
                                        realization = None,
                                        local_property_kind_uuid = None,
                                        extra_metadata = {},
                                        new_epc_file = None):
    """Adds an edges per column grid property from a numpy array to an existing resqml dataset.

    arguments:
       epc_file (string): file name to load model resqml model from (and rewrite to if new_epc_file is None)
       a (3D numpy array): the property array to be added to the model; expected shape (nj,ni,2,2) or (nj,ni,4)
       property_kind (string): the resqml property kind
       grid_uuid (uuid object or string, optional): the uuid of the grid to which the property relates;
          if None, the property is attached to the 'main' grid
       source_info (string): typically the name of a file from which the array has been read but can be any
          information regarding the source of the data
       title (string): this will be used as the citation title when a part is generated for the array; for simulation
          models it is desirable to use the simulation keyword when appropriate
       discrete (boolean, default False): if True, the array should contain integer (or boolean) data; if False, float
       uom (string, default None): the resqml units of measure for the data; not relevant to discrete data
       time_index (integer, default None): if not None, the time index to be used when creating a part for the array
       time_series_uuid (uuid object or string, default None): required if time_index is not None
       string_lookup_uuid (uuid object or string, optional): required if the array is to be stored as a categorical
          property; set to None for non-categorical discrete data; only relevant if discrete is True
       null_value (int, default None): if present, this is used in the metadata to indicate that this value
          is to be interpreted as a null value wherever it appears in the data (use for discrete data only)
       facet_type (string): resqml facet type, or None
       facet (string): resqml facet, or None
       realization (int): realization number, or None
       local_property_kind_uuid (uuid.UUID or string): uuid of local property kind, or None
       extra_metadata (dict, optional): any items in this dictionary are added as extra metadata to the new
          property
       new_epc_file (string, optional): if None, the source epc_file is extended with the new property object; if present,
          a new epc file (& associated h5 file) is created to contain a copy of the grid and the new property

    returns:
       uuid.UUID - the uuid of the newly created property

    notes:
       the RESQML protocol for saving edges per column properties uses a clockwise ordering of the 4 edges
       of a column; the resqpy protocol uses 2 dimensions of extent 2, being the axis (J, I) and face (-, +);
       this function assumes the array is in RESQML protocol if it has shape (nj, ni, 4) and resqpy protocol
       if it has shape (nj, ni, 2, 2); when reloading the property it will be presented in RESQML protocol;
       calling code can use property module functions reformat_column_edges_from_resqml_format() and
       reformat_column_edges_to_resqml_format() to convert between the protocols if needed
    """

    assert a.ndim in [3, 4]
    if a.ndim == 4:  # resqpy protocol
        assert a.shape[2] == 2 and a.shape[3] == 2, 'Wrong shape! Expected shape (nj, ni, 2, 2)'
        array_rq = rqp.reformat_column_edges_to_resqml_format(a)
    else:  # RESQML protocol
        assert a.shape[2] == 4, 'Wrong shape! Expected shape (nj, ni, 4)'
        array_rq = a

    property_uuid = rqdm_aogp.add_one_grid_property_array(epc_file,
                                                          array_rq,
                                                          property_kind,
                                                          grid_uuid = grid_uuid,
                                                          source_info = source_info,
                                                          title = title,
                                                          discrete = discrete,
                                                          uom = uom,
                                                          time_index = time_index,
                                                          time_series_uuid = time_series_uuid,
                                                          string_lookup_uuid = string_lookup_uuid,
                                                          null_value = null_value,
                                                          indexable_element = 'edges per column',
                                                          facet_type = facet_type,
                                                          facet = facet,
                                                          realization = realization,
                                                          local_property_kind_uuid = local_property_kind_uuid,
                                                          count_per_element = 1,
                                                          extra_metadata = extra_metadata,
                                                          new_epc_file = new_epc_file)
    return property_uuid
